"""
@File: binding.py
@Date: 2024/12/10 10:00
@desc: sql工具
"""


def get_insert_sql(table: str, data: dict) -> (str, list):
    """
    获取insert语句
    :param table: 表名
    :param data: insert语句的数据
    :return: 返回插入SQL, 和绑定参数
    """
    # 参数检查
    if len(data) == 0:
        return "", []

    # 构建 COLUMN 子句
    column_clause_str = ', '.join(data.keys())

    # 构建 VALUES 子句
    values_clause_str = ', '.join(['%s'] * len(data.values()))

    query = f"INSERT INTO {table} ({column_clause_str}) VALUES ({values_clause_str})"
    params = list(data.values())
    return query, params


def get_insert_or_update_sql(table: str, data: dict, update_data: dict) -> (str, list):
    """
    获取insert或update语句
    :param table: 表名
    :param data: insert语句的数据
    :param update_data: 需要更新的列名和值的字典
    :return: 返回插入或更新SQL, 和绑定参数
    """
    # 参数检查
    if len(data) == 0:
        return "", []

    # 构建 COLUMN 子句
    column_clause_str = ', '.join(data.keys())

    # 构建 VALUES 子句
    values_clause_str = ', '.join(['%s'] * len(data.values()))

    # 构建 ON DUPLICATE KEY UPDATE 子句
    update_clause_list = []
    update_params = []
    for key, value in update_data.items():
        if isinstance(value, str) and '+' in key:  # 自增表达式检查
            key = key.replace("+", "")
            update_clause_list.append(f"{key} = {key} + {value}")
        elif isinstance(value, str) and '-' in key:  # 自减表达式检查
            key = key.replace("-", "")
            update_clause_list.append(f"{key} = {key} - {value}")
        else:
            update_clause_list.append(f"{key} = %s")
            update_params.append(value)

    update_clause_str = ', '.join(update_clause_list)

    query = f"INSERT INTO {table} ({column_clause_str}) VALUES ({values_clause_str}) ON DUPLICATE KEY UPDATE {update_clause_str}"
    params = list(data.values()) + update_params
    return query, params


def get_update_sql(table: str, data: dict, where: dict) -> (str, list):
    """
    获取update语句
    :param table: 表名
    :param data: update语句的数据
    :param where: update语句的where条件
    :return: 返回更新SQL, 和绑定参数
    """
    # 参数检查
    if len(data) == 0 or len(where) == 0:
        return "", []
    params = []

    # 构建 SET 子句
    set_clause_parts = []
    for key, value in data.items():
        params.append(value)
        set_clause_parts.append(f"{key} = %s")
    set_clause_str = ', '.join(set_clause_parts)

    # 构建 WHERE 子句
    condition_clause_parts = []
    for key, value in where.items():
        comparator = "="
        for comp in ['>=', '<=', '>', '<', '!=', '<>']:
            if comp in key:
                comparator = comp
                key = key.replace(comparator, '')
                break

        if isinstance(value, list):
            # 如果值是数组，使用 IN 语句
            params.extend(value)
            condition_clause_parts.append(f"{key} IN ({', '.join(['%s'] * len(value))})")
        else:
            params.append(value)
            condition_clause_parts.append(f"{key} {comparator} %s")
    condition_clause_str = " AND ".join(condition_clause_parts)

    query = f"UPDATE {table} SET {set_clause_str} WHERE {condition_clause_str}"
    return query, params


def get_select_by_where_sql(
        table: str,
        column_list: list[str],
        where: dict,
        order_by: str = "",
        limit: int = 10000) -> (str, list):
    """
    根据条件获取SELECT查询语句
    :param table: 表名
    :param column_list: 列名
    :param where: WHERE子句字典，包含字段和对应的值
    :param order_by: 排序
    :param limit: 上限
    :return: 返回查询SQL, 和绑定参数
    """
    # 参数检查
    if len(column_list) == 0 or len(where) == 0:
        return "", []
    params = []

    # 构建 COLUMN 子句
    column_clause_str = ", ".join(column_list)

    # 构建 WHERE 子句
    condition_clause_parts = []
    for key, value in where.items():
        comparator = "="
        for comp in ['>=', '<=', '>', '<', '!=', '<>', 'LIKE', 'like']:
            if comp in key:
                comparator = comp
                key = key.replace(comparator, '').strip()
                break

        if isinstance(value, list):
            # 如果值是数组，使用 IN 语句
            params.extend(value)
            condition_clause_parts.append(f"{key} IN ({', '.join(['%s'] * len(value))})")
        else:
            params.append(value)
            condition_clause_parts.append(f"{key} {comparator} %s")
    condition_clause_str = " AND ".join(condition_clause_parts)

    # 构建 ORDER 子句
    order_by_clause_str = ""
    if order_by:
        order_by_clause_str = f"ORDER BY {order_by}"

    # 构建 SELECT 查询语句
    query = f"SELECT {column_clause_str} FROM {table} WHERE {condition_clause_str} {order_by_clause_str} LIMIT {limit};"
    return query, params



"""
@File: custom_handler.py
@Date: 2024/12/10 10:00
@desc: 第三方日志回调模块
"""
from wpylib.util.x.xjson import stringify
from typing import Dict, Any, List, Union
from langchain_core.outputs import LLMResult
from langchain_core.messages import BaseMessage
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.agents import AgentAction, AgentFinish
import logging


class LogCallback(BaseCallbackHandler):
    """
    日志回调
    接Langchain的回调: https://python.langchain.com/docs/modules/callbacks/
    """
    # 基础属性
    _name: str = ""
    _class: str = ""
    _logger: logging.Logger

    def __init__(self, name: str = "langchain日志回调"):
        self._name = name
        self._class = "LogCallback"
        self._logger = logging.getLogger("wpylib_logger")

    def on_llm_start(
            self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> Any:
        """Run when LLM starts running."""
        result_snapshot = {"serialized": serialized, "prompts": prompts}
        logger = logging.getLogger("wpylib_logger")
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_llm_start" + stringify(result_snapshot))

    def on_chat_model_start(
            self, serialized: Dict[str, Any], messages: List[List[BaseMessage]], **kwargs: Any
    ) -> Any:
        """Run when Chat Model starts running."""
        result_snapshot = {"serialized": serialized, "messages": messages}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_chat_model_start" + stringify(result_snapshot))

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""
        result_snapshot = {"token": token}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_llm_new_token" + stringify(result_snapshot))

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        result_snapshot = {"response": response}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_llm_end" + stringify(result_snapshot))

    def on_llm_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when LLM errors."""
        result_snapshot = {"error": error}
        self._logger.error("ERROR: langchain -> callback -> log_callback -> on_llm_error" + stringify(result_snapshot))

    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        """Run when chain starts running."""
        result_snapshot = {"serialized": serialized, "inputs": inputs}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_chain_start" + stringify(result_snapshot))

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        """Run when chain ends running."""
        result_snapshot = {"outputs": outputs}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_chain_end" + stringify(result_snapshot))

    def on_chain_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when chain errors."""
        result_snapshot = {"error": error}
        self._logger.error("ERROR: langchain -> callback -> log_callback -> on_chain_error" + stringify(result_snapshot))

    def on_tool_start(
            self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        """Run when tool starts running."""
        result_snapshot = {"serialized": serialized, "input_str": input_str}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_tool_start" + stringify(result_snapshot))

    def on_tool_end(self, output: Any, **kwargs: Any) -> Any:
        """Run when tool ends running."""
        result_snapshot = {"output": output}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_tool_end" + stringify(result_snapshot))

    def on_tool_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when tool errors."""
        result_snapshot = {"error": error}
        self._logger.error("ERROR: langchain -> callback -> log_callback -> on_tool_error" + stringify(result_snapshot))

    def on_text(self, text: str, **kwargs: Any) -> Any:
        """Run on arbitrary text."""
        result_snapshot = {"text": text}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_text" + stringify(result_snapshot))

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        """Run on agent action."""
        result_snapshot = {"action": action}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_agent_action" + stringify(result_snapshot))

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> Any:
        """Run on agent end."""
        result_snapshot = {"finish": finish}
        self._logger.info("INFO: langchain -> callback -> log_callback -> on_agent_finish" + stringify(result_snapshot))



"""
@File: cmd.py
@Date: 2024/12/10 10:00
@desc: 命令行工具
"""
import sys


def args_to_dict():
    """
    将系统参数转为KV结构的kwargs
    """
    if len(sys.argv) <= 1:
        return []

    kwargs = {}
    for arg in sys.argv[0:]:
        chunks = arg.split("=")
        if len(chunks) < 2:
            continue

        pieces = chunks[0].split("--")
        if len(pieces) < 2:
            continue

        k = pieces[1]
        v = chunks[1]
        kwargs[k] = v
    return kwargs



"""
@File: xtyping.py
@Date: 2024/12/10 10:00
@desc: 类型工具
"""


def is_none(data):
    """
    数据是否是None
    :param data: 数据
    :return 是否为None
    """
    if data is None:
        return True
    return False


def is_not_none(data):
    """
    数据是否不是None
    :param data: 数据
    :return 是否不为None
    """
    return is_none(data) is False



"""
@File: xjson.py
@Date: 2024/12/10 10:00
@Desc: JSON工具
"""
import json
import re


def _remove_markdown_of_json(raw_json: str):
    """
    去除markdown的JSON标签
    :param raw_json: 原JSON
    :return: 返回去除后的字符串
    """
    raw_json = raw_json.replace("```json", '')
    raw_json = raw_json.replace("```JSON", '')
    raw_json = raw_json.replace("```", '')
    return raw_json


def parse(raw_json: str, support_markdown: bool = True) -> (bool, dict):
    """
    解析数据为JSON
    :param raw_json: json字符串
    :param support_markdown: 是否支持markdown
    :return 返回是否解析成功以及解析后的JSON字典
    """
    # 如果支持markdown
    if support_markdown:
        raw_json = _remove_markdown_of_json(raw_json)

    # 直接尝试JSON解析
    try:
        data = json.loads(raw_json)
        return True, data
    except Exception as e:
        return False, None


def parse_raise(raw_json: str, support_markdown: bool = True) -> dict:
    """
    解析数据为JSON, 如果解析失败则直接抛出异常
    :param raw_json: json字符串
    :param support_markdown: 是否支持markdown
    :return 返回解析后的JSON字典
    """
    # 如果支持markdown
    if support_markdown:
        raw_json = _remove_markdown_of_json(raw_json)

    # 直接尝试JSON解析
    try:
        data = json.loads(raw_json)
        return data
    except Exception as e:
        raise RuntimeError(e)


def convert_to_jsonable(data):
    """
    强制转换为一定可以JSON编码的数据
    比如一个dict中如果有time、对象等原生类型, 那么是不可以被JSON编码的
    所以这个函数会递归遍历每一个属性, 只要不是JSON可以编码和解析的, 就会使用str强制转换
    :param data: 原数据
    :return: 返回处理后的数据
    """
    if isinstance(data, dict):
        return {k: convert_to_jsonable(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [convert_to_jsonable(item) for item in data]
    elif not isinstance(data, (int, float, str, bool, type(None))):
        # 对于非基本类型，尝试将其转换为字符串
        return str(data)
    else:
        return data


def stringify(data) -> str:
    """
    编码为JSON
    :param data: 被编码的数据
    :return string
    """
    return json.dumps(convert_to_jsonable(data), ensure_ascii=False)


def extract_first_json(text: str) -> dict:
    """
    提取字符串中的第一个```json```代码块内的JSON内容
    :param text: 原字符串
    :return: 返回解析后的JSON字典
    """
    match = re.search(r'```json\s*(.*?)\s*```', text, re.DOTALL)

    if match:
        json_content = match.group(1)
        return parse_raise(json_content)
    return {}



"""
@File: validate_post_params.py
@Date: 2024/1/13 14:07
@desc: 验证POST请求参数中间件
"""
from functools import wraps
from flask import request, g
from wpylib.util.http import resp_error, get_params


def validate_post_params_middleware(form_class):
    """
    验证参数
    """

    def decorator(func):
        """
        注解
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            wrapper
            """
            # 获取URL和请求参数
            url_params, post_params = get_params()
            form = form_class(meta={'csrf': False}, data=post_params)

            # 参数验证失败
            if not form.validate():
                # 返回错误信息
                return resp_error(msg="参数错误", code=4001)

            # 参数验证成功
            if "context_data" not in g:
                g.context_data = {}
            g.context_data["form_info"] = form.data
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""
@File: check_login_stream.py
@Date: 2024/1/13 14:07
@desc: 流式接口中检查登录中间件
"""
from functools import wraps
from flask import request, g
import threading


def check_login_stream_middleware(func):
    """
    检查登录中间件
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """
        注解wrapper
        """

        # ...

        # 登录成功
        user = {"userId": 99}
        if "context_data" not in g:
            g.context_data = {}
        g.context_data["login_info"] = user
        g.context_data["login_info"]["user_id"] = user["userId"]
        g.context_data["login_info"]["userid"] = user["userId"]
        g.context_data["login_info"]["userId"] = user["userId"]

        thread_local = threading.local()
        setattr(thread_local, "wpylib_threadkey_user_id", user["userId"])

        # 放行
        return func(*args, **kwargs)

    return wrapper



"""
@File: encry.py
@Date: 2024/12/10 10:00
@desc: 编码加密工具
"""
import uuid
import hashlib


def sha1(input_string: str) -> str:
    """
    sha1加密
    :param input_string: 输入的字符串
    :return: 返回使用sha1算法计算后的加密字符串
    """
    # 创建sha1哈希函数对象
    sha1_hash_func = hashlib.sha1()

    # 计算输入字符串的哈希值
    sha1_hash_func.update(input_string.encode('utf-8'))

    # 获取哈希值的十六进制值
    hashed_string = sha1_hash_func.hexdigest()

    # 返回十六进制的哈希值作为计算后的加密字符串
    return hashed_string


def gen_random_md5(node=0x87654321) -> str:
    """
    生成随机md5字符串
    :param node: 使用uuid1生成唯一ID时的节点信息
    :return: 返回生成的md5字符串
    """
    log_id = uuid.uuid1(node)
    return log_id.hex.replace('-', '')



"""
@File: stream_queue.py
@Date: 2024/12/10 10:00
@Desc: 第三方队列模块
"""
from wpylib.util.x.xtyping import is_none
import queue
import time


class StreamQueue(queue.Queue):
    """
    StreamQueue流式队列
    """

    def __init__(self, maxsize=0):
        super().__init__(maxsize)
        self.running = True

    def close(self):
        """
        关闭队列
        :return:
        """
        self.join()
        self.running = False

    def is_running(self):
        """
        是否正在运行
        :return:
        """
        return self.running

    def put(self, item, block=True, timeout=None):
        """
        阻塞放入队列
        :param item: 队列元素
        :param block: 是否阻塞
        :param timeout: 超时时间
        :return:
        """
        raise RuntimeError("Stream cannot use put method, using send_message or send_message_end instead")

    def put_nowait(self, item):
        """
        非阻塞放入队列
        :param item: 队列元素
        :return:
        """
        raise RuntimeError(
            "Stream cannot use put_nowait method, using send_message_nowait or send_message_end_nowait instead"
        )

    def send_message(self, type_str: str, item: dict = None, name="", is_show=True, block=True, timeout=None) -> float:
        """
        阻塞发送消息
        :param type_str: 消息类型
        :param item: 队列元素
        :param name: 名称
        :param is_show: 前端是否展示
        :param block: 是否阻塞
        :param timeout: 超时时间
        :return:
        """
        if not self.running:
            raise RuntimeError("stream queue already closed.")
        if is_none(item):
            item = {}

        # 发送消息
        data = {
            "event": "message",
            "data": {
                "type": type_str,
                "name": name,
                "is_show": is_show,
                "item": item
            }
        }
        super().put(item=data, block=block, timeout=timeout)
        return time.time()

    def send_message_end(self, data: dict, block=True, timeout=None) -> float:
        """
        阻塞发送结束消息
        :param data: 消息内容
        :param block: 是否阻塞
        :param timeout: 超时时间
        :return:
        """
        if not self.running:
            raise RuntimeError("stream queue already closed.")
        super().put(item={"event": "message_end", "data": data}, block=block, timeout=timeout)
        self.close()
        return time.time()


class NoneQueue(StreamQueue):
    """
    空队列
    """

    def send_message(self, type_str: str, item: dict = None, name="", is_show=True, block=True, timeout=None) -> float:
        return time.time()

    def send_message_end(self, data: dict, block=True, timeout=None):
        return time.time()



"""
@File: model.py
@Date: 2024/12/10 10:00
@desc: 第三方model单例模块
"""
from typing import Any, Optional, Iterator
from langchain_openai import OpenAI, AzureOpenAI
from langchain_core.runnables.utils import Input, Output
from langchain_openai import ChatOpenAI, AzureChatOpenAI
from langchain_core.runnables.config import RunnableConfig
from langchain_core.language_models.base import BaseLanguageModel
from langchain_community.chat_models import ChatZhipuAI, ChatBaichuan

# 角色
MODEL_ROLE_SYSTEM = "SYSTEM"
MODEL_ROLE_HUMAN = "HUMAN"
MODEL_ROLE_AI = "AI"

# MODEL类型
MODEL_TYPE = str
MODEL_TYPE_OPENAI_STRING: MODEL_TYPE = "model_type_openai_string"
MODEL_TYPE_OPENAI_CHAT: MODEL_TYPE = "model_type_openai_chat"
MODEL_TYPE_AZURE_STRING: MODEL_TYPE = "model_type_azure_string"
MODEL_TYPE_AZURE_CHAT: MODEL_TYPE = "model_type_azure_chat"
MODEL_TYPE_ZHIPU_CHAT: MODEL_TYPE = "model_type_zhipu_chat"
MODEL_TYPE_BAICHUAN_CHAT: MODEL_TYPE = "model_type_baichuan_chat"
MODEL_TYPE_DOUBAO_CHAT: MODEL_TYPE = "model_type_doubao_chat"

# Embedding类型
EMBEDDING_TYPE = str
EMBEDDING_TYPE_OPENAI: EMBEDDING_TYPE = "embedding_type_openai"
EMBEDDING_TYPE_AZURE: EMBEDDING_TYPE = "embedding_type_azure"
EMBEDDING_TYPE_HUGGING_FACE: EMBEDDING_TYPE = "embedding_type_hugging_face"


class Model:
    """
    Model模型
    """
    # 基础属性
    _model_type: MODEL_TYPE = MODEL_TYPE_AZURE_CHAT

    # 使用到的实例对象
    _raw_model: BaseLanguageModel

    def __init__(
            self,
            model_type: MODEL_TYPE = MODEL_TYPE_AZURE_CHAT,
            model_config: dict = None,
            temperature=0.5,
            cost_strict: bool = False,
            **kwargs
    ):
        # 判断是否开启cost_strict: 严格消费控制, 免得调用未经费用确认的model, 造成费用开销
        if cost_strict and model_type != MODEL_TYPE_AZURE_CHAT:
            raise RuntimeError(f"only support AZURE_CHAT model_type: {model_type}")
        self._model_type = model_type

        # 根据不同类型创建不同的Model
        if model_type == MODEL_TYPE_OPENAI_CHAT:
            self._raw_model = ChatOpenAI(
                temperature=temperature,
                max_retries=model_config["retry"],
                **kwargs
            )
        elif model_type == MODEL_TYPE_DOUBAO_CHAT:
            self._raw_model = ChatOpenAI(
                openai_api_key=model_config["api_key"],
                openai_api_base=model_config["api_base"],
                model_name=model_config["model"],
                temperature=temperature,
                max_retries=model_config["retry"],
                **kwargs
            )
        elif model_type == MODEL_TYPE_OPENAI_STRING:
            self._raw_model = OpenAI(
                temperature=temperature,
                max_retries=model_config["retry"],
                **kwargs
            )
        elif model_type == MODEL_TYPE_AZURE_CHAT:
            self._raw_model = AzureChatOpenAI(
                temperature=temperature,
                openai_api_key=model_config["api_key"],
                azure_endpoint=model_config["azure_endpoint"],
                openai_api_version=model_config["api_version"],
                model=model_config["model_version"],
                max_retries=model_config["retry"],
                **kwargs
            )
        elif model_type == MODEL_TYPE_AZURE_STRING:
            self._raw_model = AzureOpenAI(
                temperature=temperature,
                openai_api_key=model_config["api_key"],
                azure_endpoint=model_config["azure_endpoint"],
                openai_api_version=model_config["api_version"],
                model=model_config["model_version"],
                max_retries=model_config["retry"],
                **kwargs
            )
        elif model_type == MODEL_TYPE_BAICHUAN_CHAT:
            # 目前社区百川封装的模块, 暂不支持这种使用
            # 只能暂时先使用调接口的方式, 见BaichuanHTTPNode
            self._raw_model = ChatBaichuan(
                baichuan_api_key=model_config["api_key"],
                baichuan_api_base=model_config["api_base"],
                model=model_config["model"],
                temperature=temperature,
                max_retries=model_config["retry"],
                **kwargs
            )
        elif model_type == MODEL_TYPE_ZHIPU_CHAT:
            max_tokens = 1024
            if "max_tokens" in model_config:
                max_tokens = model_config["max_tokens"]
            self._raw_model = ChatZhipuAI(
                zhipuai_api_key=model_config["api_key"],
                zhipuai_api_base=model_config["api_base"],
                model=model_config["model"],
                temperature=temperature,
                max_retries=model_config["retry"],
                max_tokens=max_tokens,
                **kwargs
            )
        else:
            raise RuntimeError(f"unknown model_type: {model_type}")

    def stream(
            self,
            langchain_input: Input,
            config: Optional[RunnableConfig] = None,
            **kwargs: Optional[Any],
    ) -> Iterator[Output]:
        """
        流式输出
        :param langchain_input:
        :param config:
        """
        return self._raw_model.stream(langchain_input, config, **kwargs)

    def get_raw_model(self) -> BaseLanguageModel:
        """
        返回实例model
        """
        return self._raw_model



"""
@File: localcache.py
@Date: 2024/12/10 10:00
@Desc: 第三方本地缓存单例模块
"""
from typing import Any
from wpylib.util.x.xtyping import is_none
import threading


class Localcache:
    """
    Localcache本地缓存类
    """
    # 本地线程存储
    _thread_local: Any

    # 单例类
    _singleton = None

    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
        return cls._singleton

    def __init__(self):
        self._thread_local = threading.local()

    def set_thread_local(self, key: Any, value: Any):
        """
        设置thread_local的Key-Value对
        :param key: 缓存的Key
        :param value: 缓存的值
        :return:
        """
        setattr(self._thread_local, key, value)

    def get_thread_local(self, key: Any) -> Any:
        """
        根据Key获取thread_local的Value
        :param key: 缓存的Key
        :return:
        """
        try:
            return getattr(self._thread_local, key)
        except Exception as e:
            return None

    def set_log_id(self, value: str):
        """
        设置log_id, log_id是程序运行过程中分配的唯一日志ID
        :param value: log_id的值
        :return:
        """
        key = "wpylib_threadkey_log_id"
        self.set_thread_local(key, value)

    def get_log_id(self) -> str:
        """
        获取log_idd的值
        :return:
        """
        key = "wpylib_threadkey_log_id"
        log_id = self.get_thread_local(key)
        if is_none(log_id):
            log_id = ""
        return log_id

    def set_user_id(self, value: int):
        """
        设置user_id
        :param value: user_id的值
        :return:
        """
        key = "wpylib_threadkey_user_id"
        self.set_thread_local(key, value)

    def get_user_id(self) -> int:
        """
        获取user_id
        :return: 返回user_id
        """
        key = "wpylib_threadkey_user_id"
        user_id = self.get_thread_local(key)
        if is_none(user_id):
            user_id = 0
        return user_id




"""
@File: simple_web_document.py
@Date: 2024/2/27 00:07
@Desc: WEB文档结构
"""
from typing import Any
from langchain_core.documents import Document as LangChainDocument


class SimpleWebDocument(LangChainDocument):
    """
    Web文档对象
    """
    # 注意: 如果使用私有属性, 会提示没有此属性, 所以修改为公开属性
    title: str = ""
    url: str = ""
    description: str = ""
    content: str = ""

    def __init__(self, title: str = "", url: str = "", description: str = "", content: str = "", **kwargs: Any):
        super().__init__(content, **kwargs)
        self.title = title
        self.url = url
        self.description = description
        self.content = content

    def get_title(self) -> str:
        """
        获取title
        """
        return self.title

    def get_url(self) -> str:
        """
        获取url
        """
        return self.url

    def get_description(self) -> str:
        """
        获取description
        """
        return self.description

    def get_content(self) -> str:
        """
        获取content
        """
        return self.content



"""
@File: milvus.py
@Date: 2024/12/10 10:00
@Desc: milvus向量存储模块
"""
from pymilvus import MilvusClient
from wpylib.util.x.xtyping import is_none
from mem0.utils.factory import EmbedderFactory
from typing import Union, Dict, List, Optional


class Milvus:
    """
    milvus对象
    """
    # 初始化需要的配置
    _milvus_config: dict

    # 使用到的实例对象
    _instance_milvus: MilvusClient

    # 单例类
    _singleton = None

    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
        return cls._singleton

    def __init__(self, milvus_config: dict, model_config: dict):
        self._milvus_config = milvus_config

        # 初始化embedding模型
        self._embedding_model = EmbedderFactory.create(
            provider_name="azure_openai",
            config={
                "model": model_config["model"],  # 如text-embedding-ada-002
                "azure_kwargs": {
                    "api_version": model_config["api_version"],  # 如2024-01-03-preview
                    "azure_deployment": model_config["azure_deployment"],  # 如text-embedding-ada-002
                    "azure_endpoint": model_config["azure_endpoint"],  # 如https://xxx.openai.azure.com/
                    "api_key": model_config["api_key"]  # 如3e6f8...
                }
            }
        )

        # 初始化milvus客户端
        self._instance_milvus = MilvusClient(
            uri=milvus_config["uri"],
            # token=milvus_config["token"],
            db_name=milvus_config["db_name"],
            # ssl=False,
            # secure=False
        )
        for collection_name in milvus_config["collection"]:
            self._instance_milvus.load_collection(collection_name=collection_name)  # 加载collection, 否则可能会报错

        # 验证milvus客户端是否可用
        test_collection_name = milvus_config["collection"]["test"]
        search_res = self.search(
            collection_name=test_collection_name, query="test", output_fields=["*"]
        )

    def get_instance_milvus(self) -> MilvusClient:
        """
        获取实例
        :return:
        """
        return self._instance_milvus

    def embed(self, text: str) -> list[float]:
        """
        文本向量化
        :param text:
        :return:
        """
        embeddings = self._embedding_model.embed(text)
        return embeddings

    def query(
            self,
            collection_name: str,
            filter_str: str = "",
            output_fields: Optional[List[str]] = None,
            timeout: Optional[float] = None,
            ids: Optional[Union[List, str, int]] = None,
            partition_names: Optional[List[str]] = None
    ):
        """
        查询记录
        :param collection_name:
        :param filter_str:
        :param output_fields:
        :param timeout:
        :param ids:
        :param partition_names:
        :return:
        """
        query_res = self._instance_milvus.query(
            collection_name=collection_name,
            filter=filter_str,
            output_fields=output_fields,
            timeout=timeout,
            ids=ids,
            partition_names=partition_names
        )
        return query_res

    def insert(
            self,
            collection_name: str,
            data: Union[Dict, List[Dict]],
            timeout: Optional[float] = None,
            partition_name: Optional[str] = "",
    ) -> dict:
        """
        插入数据
        :param collection_name:
        :param data:
        :param timeout:
        :param partition_name:
        :return:
        """
        res = self._instance_milvus.insert(
            collection_name=collection_name,
            data=data,
            timeout=timeout,
            partition_name=partition_name
        )
        return res

    def upsert(
            self,
            collection_name: str,
            data: Union[Dict, List[Dict]],
            timeout: Optional[float] = None,
            partition_name: Optional[str] = "",
    ) -> dict:
        """
        插入或更新数据
        不是这样: 适用场景, 如果有update同一个question的需求, 是不是我先query出来(filter这个question字段), 拿出来ID后, 再使用upsert
        是这样: 如果是使用了auto_id的collection的话是不能用upsert的，您可以先delete然后再insert新的，因为auto_id是有唯一性的。实际上upsert的底层就是封装了delete和insert这两个操作，为的是代码方便
        :param collection_name:
        :param data:
        :param timeout:
        :param partition_name:
        :return:
        """
        res = self._instance_milvus.upsert(
            collection_name=collection_name,
            data=data,
            timeout=timeout,
            partition_name=partition_name
        )
        return res

    def search(
            self,
            collection_name: str,
            query: str,
            output_fields: list[str],
            search_params=None,
            filter_str: str = "",
            limit=3
    ):
        """
        搜索
        filter参数: https://milvus.io/docs/boolean.md
        search参数: https://docs.zilliz.com.cn/reference/python/python/Vector-search
        返回结构: data: ["[{'id': 452424477291365021, 'distance': 0.9999998211860657, 'entity': {}}]"]
        :param collection_name:
        :param query:
        :param output_fields:
        :param search_params:
        :param filter_str:
        :param limit:
        :return:
        """
        # 参数处理
        if is_none(search_params):
            search_params = {}
        default_search_params = {
            # 度量向量embedding间相似度的方法: 可选值为 IP、L2、COSINE。默认按已加载（load）的索引文件设定。
            "metric_type": "COSINE",
            # query failed: metric type not match: invalid parameter[expected=COSINE][actual=IP]
            "params": {}
        }
        default_search_params.update(search_params)

        # 向量化
        embeddings = self.embed(query)

        # 开始搜索
        search_list = self._instance_milvus.search(
            collection_name=collection_name,
            data=[embeddings],
            output_fields=output_fields,
            search_params=default_search_params,
            filter=filter_str,
            limit=limit
        )

        # 返回结果
        if len(search_list) > 0:
            return search_list[0]
        return []



"""
@File: mysql.py
@Date: 2024/12/10 10:00
@Desc: 第三方mysql存储单例模块
"""
from wpylib.util.x.xtyping import is_not_none
from mysql.connector.pooling import MySQLConnectionPool
import logging


# 查询语句类型
QUERY_TYPE_RAW: str = "raw"
QUERY_TYPE_INSERT: str = "insert"
QUERY_TYPE_BATCH_INSERT: str = "batch_insert"
QUERY_TYPE_DELETE: str = "delete"
QUERY_TYPE_SELECT: str = "select"
QUERY_TYPE_UPDATE: str = "update"
QUERY_TYPE_INSERT_OR_UPDATE: str = "insert_or_update"


class Mysql:
    """
    Mysql数据库类
    """
    # 基础属性
    _pool: MySQLConnectionPool

    # 初始化需要的配置
    _mysql_config: dict

    # 单例类
    _singleton = None

    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
        return cls._singleton

    def __init__(self, mysql_config: dict):
        self._mysql_config = mysql_config
        self._connect()

    def _connect(self):
        """
        连接数据库
        1. buffered=True, fix error: raise InternalError("Unread result found")
           https://stackoverflow.com/questions/29772337/python-mysql-connector-unread-result-found-when-using-fetchone
        """
        for times in range(self._mysql_config["retry"]):
            try:
                self._pool = MySQLConnectionPool(pool_name="mysql_pool", **self._mysql_config["dsn"])
                self._log("connected to mysql successfully")
                break
            except Exception as e:
                self._log("connect to mysql retry", error=True, extra_info={"retry": times, "e": e})
                if times == self._mysql_config["retry"] - 1:
                    raise RuntimeError(e)

    def _re_connect(self):
        """
        连接失效后的重连
        """
        self._connect()

    def _execute(self, query_type: str, query: str, params: list = None):
        """
        执行语句
        :param query: 查询语句
        :param params: 参数列表
        :param query_type: Query类型
        :return: 返回结果
        """
        cnx = self._pool.get_connection()
        cursor = cnx.cursor(buffered=True)
        exception = None

        # 开始执行
        result = None
        try:
            # 执行Query
            for times in range(self._mysql_config["retry"]):
                try:
                    # 单条执行
                    if query_type != QUERY_TYPE_BATCH_INSERT:
                        cursor.execute(query, params)
                    # 多条执行
                    else:
                        cursor.executemany(query, params)
                    break
                except Exception as e:
                    # 记录日志
                    self._log(
                        "execute error",
                        error=True,
                        extra_info={"retry": times, "query": query, "params": params, "e": e}
                    )
                    # 超过最大重试次数则上抛错误
                    if times == self._mysql_config["retry"] - 1:
                        raise e

            # 提交Query
            # 调试的时候需要以这个为断点, 不然会一直报连接不上Mysql的问题
            cnx.commit()

            # 根据语句类型不同, 获取不同的返回值
            query_type = query_type.lower()
            if query_type == QUERY_TYPE_INSERT or query_type == QUERY_TYPE_BATCH_INSERT:
                # 无论单条插入还是批量插入, 都只返回最后一条插入记录ID
                result = cursor.lastrowid
            elif query_type == QUERY_TYPE_INSERT_OR_UPDATE:
                # 如果是insert_or_update, 只返回最后一条插入记录ID
                result = cursor.lastrowid
            elif query_type == QUERY_TYPE_DELETE:
                # 返回影响行数
                result = cursor.rowcount
            elif query_type == QUERY_TYPE_SELECT:
                # 返回查询结果
                # 无论是查询一条记录还是多条记录, 都返回一个数组, 需要上层自己处理
                columns = [desc[0] for desc in cursor.description]
                results = cursor.fetchall()
                result = [dict(zip(columns, row)) for row in results]
            elif query_type == QUERY_TYPE_UPDATE:
                # 返回影响行数
                result = cursor.rowcount
            else:
                # RAW语句
                result = cursor.rowcount

            # 执行成功
            self._log("raw execute success", extra_info={"query": query, "params": params})
        except Exception as e:
            exception = e
            self._log("raw execute exception finally", error=True,
                      extra_info={"query": query, "params": params, "e": e})
        finally:
            # 关闭连接
            if cnx:
                cnx.close()
            # 关闭cursor
            if cursor:
                cursor.close()
            self._log("put connection to pool")

        # 处理执行结果
        if is_not_none(exception):
            raise exception
        return result

    def _log(self, message, error=False, extra_info=None):
        """
        辅助函数用于统一处理日志记录。
        :param message: 日志消息主体。
        :param error: 是否为错误信息，默认为False。
        :param extra_info: 额外信息字典，可选。
        """
        if extra_info:
            for key, value in extra_info.items():
                message += f" {key}: {value}"
        logger = logging.getLogger("wpylib_logger")
        if error:
            logger.error(message)
        else:
            logger.info(message)

    def execute_raw_query(self, query: str, params: list = None):
        """
        执行原生语句
        :param query: 语句
        :param params: 参数列表
        :return 返回类型不确定
        """
        return self._execute(QUERY_TYPE_RAW, query, params)

    def execute_insert_query(self, query: str, params: list = None) -> int:
        """
        执行insert语句
        :param query: 语句
        :param params: 参数列表
        :return 返回插入记录的ID
        """
        return self._execute(QUERY_TYPE_INSERT, query, params)

    def execute_insert_or_update_query(self, query: str, params: list = None) -> int:
        """
        执行insert_or_update语句
        :param query: 语句
        :param params: 参数列表
        :return 返回插入或更新的ID
        """
        return self._execute(QUERY_TYPE_INSERT_OR_UPDATE, query, params)

    def execute_batch_insert_query(self, query: str, params: list = None) -> int:
        """
        执行批量insert语句
        :param query: 语句
        :param params: 参数列表
        :return 返回插入的自增ID
        """
        return self._execute(QUERY_TYPE_BATCH_INSERT, query, params)

    def execute_delete_query(self, query: str, params: list = None) -> int:
        """
        执行delete语句
        :param query: 语句
        :param params: 参数列表
        :return 返回删除记录的ID
        """
        return self._execute(QUERY_TYPE_DELETE, query, params)

    def execute_select_query(self, query: str, params: list = None) -> list:
        """
        执行select语句
        :param query: 语句
        :param params: 参数列表
        :return 返回查到的列表
        """
        return self._execute(QUERY_TYPE_SELECT, query, params)

    def execute_update_query(self, query: str, params: list = None) -> int:
        """
        执行update语句
        :param query: 语句
        :param params: 参数列表
        :return 返回更新记录的ID
        """
        return self._execute(QUERY_TYPE_UPDATE, query, params)



"""
@File: xtime.py
@Date: 2024/12/10 10:00
@Desc: 时间工具
"""
from datetime import datetime
import pytz
import time

# 时间格式化串
FORMAT_STR_HUMAN: str = "%Y-%m-%d %H:%M:%S"
FORMAT_STR_COMPACT: str = "%Y%m%d%H%M%S"
SECOND = 1
MINUTE = 60 * SECOND


def now_time(time_str: str = None, format_str: str = "%Y-%m-%d %H:%M:%S") -> datetime:
    """
    获取当前的时间对象(北京时区)
    :param time_str: 当前时间字符串, 如2022-10-12 12:20:30
    :param format_str: 时间格式定义
    :return 时间对象
    """
    tz = pytz.timezone('Asia/Shanghai')
    if time_str:
        return datetime.strptime(time_str, format_str).astimezone(tz)
    return datetime.now(tz)


def now_time_str(time_str: str = None, format_str: str = "%Y-%m-%d %H:%M:%S") -> str:
    """
    获取现在的时间字符串
    :param time_str: 当前时间字符串, 如2022-10-12 12:20:30
    :param format_str: 时间格式定义
    :return 时间字符串
    """
    return now_time(time_str, format_str).strftime(format_str)


def now_timestamp(time_str: str = None) -> int:
    """
    获取现在的时间戳数字
    :param time_str: 当前时间字符串, 如2022-10-12 12:20:30
    :return 时间戳
    """
    return int(datetime.timestamp(now_time(time_str)))


def now_date_str(time_str: str = None, format_str: str = "%Y-%m-%d") -> str:
    """
    获取现在的日期字符串
    :param time_str: 当前时间字符串, 如2022-10-12 12:20:30
    :param format_str: 时间格式定义
    :return 时间字符串
    """
    return now_time(time_str).strftime(format_str)


def beijing_timetuple(sec, what) -> time.struct_time:
    """
    获取北京时间的时间元组
    :return 时间元组
    """
    return now_time().timetuple()



"""
@File: access.py
@Date: 2024/1/13 14:07
@desc: access中间件
"""
from functools import wraps
from flask import request, g
from wpylib.util.x.xjson import stringify
from wpylib.util.encry import gen_random_md5
from wpylib.util.http import get_params, get_headers
import threading
import logging


def access_middleware(func):
    """
    access_middleware中间件
    :param func: 被装饰的函数
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        """
        注解wrapper
        """
        # 请求参数处理
        url_params, post_params = get_params()

        # 生成Log_id
        # Log_id由上层服务生成, 比如网关为每个请求附加一个唯一的log_id
        log_id = get_headers("x-logid")
        if log_id == "":
            log_id = gen_random_md5()
        thread_local = threading.local()
        setattr(thread_local, "wpylib_threadkey_log_id", log_id)

        # 记录日志
        logger = logging.getLogger("wpylib_logger")
        logger.info("INFO: " + stringify({
            "data": {
                "msg": "access_middleware",
                "post_params": post_params,
                "url_params": url_params,
                "url": request.url,
                "headers": request.headers,
                "log_id": log_id,
            }
        }))

        # 获取启动信息
        if "context_data" not in g:
            g.context_data = {}
        g.context_data["start_info"] = {
            "log_id": log_id
        }
        return func(*args, **kwargs)

    return wrapper



"""
@File: prompt.py
@Date: 2024/12/10 10:00
@Desc: OpenAI提供了chat和非chat的两种API, chat:/chat/completions, 非chat: /completions
@Desc: 所以区别chat还是非chat, 判断URL接口中是否有/chat/completions, 有则是Chat的API, 否则是非Chat的API
"""
from langchain.prompts import PromptTemplate
from langchain_core.prompts.chat import ChatPromptTemplate
from langchain_core.prompts.base import BasePromptTemplate
from langchain_core.prompts.chat import SystemMessage, HumanMessagePromptTemplate

# PROMPT类型
PROMPT_TYPE = str
PROMPT_TYPE_STRING: PROMPT_TYPE = "string_prompt"
PROMPT_TYPE_CHAT: PROMPT_TYPE = "chat_prompt"


def create_string_prompt(template: str = "", **kwargs) -> BasePromptTemplate:
    """
    创建文本类Prompt
    :param template: 提示词模板
    :param kwargs:
    :return:
    """
    return PromptTemplate.from_template(template, **kwargs)


def create_chat_prompt_by_messages(messages: list = None, **kwargs) -> BasePromptTemplate:
    """
    使用消息列表, 创建聊天类Prompt
    :param messages: 消息列表
    :param kwargs:
    :return:
    """
    return ChatPromptTemplate.from_messages(
        messages,
        **kwargs
    )


def create_chat_prompt_by_system_content(prompt_system_content: str, **kwargs) -> BasePromptTemplate:
    """
    使用系统提示词, 创建聊天类prompt
    :param prompt_system_content: 系统提示词部分内容
    :param kwargs:
    :return:
    """
    # (1) 拼装messages
    messages = [
        SystemMessage(prompt_system_content),
        HumanMessagePromptTemplate.from_template("{user_input_placeholder}")
    ]
    # (2) 创建prompt实例
    return ChatPromptTemplate.from_messages(messages, **kwargs)



"""
@File: web_loader.py
@Date: 2024/2/27 00:07
@Desc: 第三方Web文档加载器
"""
from typing import List
from wpylib.util.http import requests_get
from wpylib.pkg.singleton.loader.simple_web_document import SimpleWebDocument

WEB_LOADER_ENGINE_JINA = "jina"


class WebLoader:
    """
    Web文档加载器
    """
    # 变量
    _engine: str = WEB_LOADER_ENGINE_JINA
    _jina_engine_config: dict = {
        "host": "https://r.jina.ai",
        "headers": {
            "Accept": "application/json",
        }
    }

    def __init__(self, engine: str = WEB_LOADER_ENGINE_JINA):
        self._engine = engine

    def load(self, resource_info_list: list[dict], proxies: dict = None, requests_kwargs: dict = None,
             headers: dict = None) -> List[SimpleWebDocument]:
        """
        加载网页
        """
        new_docs: List[SimpleWebDocument] = []

        # JINA爬虫引擎
        if self._engine == WEB_LOADER_ENGINE_JINA:
            new_docs = self._load_by_jina(
                resource_info_list=resource_info_list,
                proxies=proxies,
                requests_kwargs=requests_kwargs,
                headers=headers
            )
        return new_docs

    def _load_by_jina(self, resource_info_list: list[dict], proxies: dict = None, requests_kwargs: dict = None,
                      headers: dict = None) -> List[SimpleWebDocument]:
        """
        使用jina加载网页, jina控制参数都是在header中设置的
        """
        # 参数处理
        if requests_kwargs is None:
            requests_kwargs = {}
        if headers is None:
            headers = {}
        headers = dict(self._jina_engine_config["headers"], **headers)

        # 开始准备爬取网页
        new_docs: List[SimpleWebDocument] = []
        for url_info in resource_info_list:
            url = self._jina_engine_config["host"] + "/" + url_info["url"]
            resp = requests_get(url=url, headers=headers)
            new_docs.append(SimpleWebDocument(
                title=resp["data"]["title"],
                url=resp["data"]["url"],
                description=resp["data"]["description"],
                content=resp["data"]["content"],
            ))

        # 生成新的文档
        return new_docs



"""
@File: logger.py
@Date: 2024/12/10 10:00
@Desc: 第三方logger单例模块
"""
from wpylib.util.x.xjson import convert_to_jsonable
from wpylib.util.storage import create_directory, is_directory_exist
from wpylib.util.x.xtime import now_timestamp, now_time_str, now_date_str, beijing_timetuple
import traceback
import threading
import logging
import inspect
import json
import sys


class Logger:
    """
    二次封装的Logger日志类
    """
    # 初始化需要的配置
    _logger_config: dict
    _global_config: dict

    # 原生logger实例
    _instance_logger: logging.Logger

    # Logger类的单例
    _singleton = None

    def __new__(cls, *args, **kwargs):
        # 在初始化Logger类的时候, 确保仅存在一个单例
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
        return cls._singleton

    def __init__(self, logger_config: dict, global_config: dict):
        if "level" not in logger_config:
            logger_config["level"] = "info"
        if "log_format" not in logger_config:
            logger_config["log_format"] = '%(asctime)s - %(message)s'
        self._logger_config = logger_config

        if "global_base_dir" not in global_config:
            global_config["global_base_dir"] = "./"
        if "project" not in global_config:
            global_config["project"] = ""
        if "business" not in global_config:
            global_config["business"] = ""
        self._global_config = global_config

        self._init()

    def _init(self):
        """
        初始化Logger
        """
        # 日志初始配置
        date = now_date_str(None, "%Y%m%d")
        logging.Formatter.converter = beijing_timetuple

        # 设置全局日志级别, 日志优先级为 ERROR > INFO > DEBUG
        # 由于INFO日志优先级高于DEBUG，INFO级别的消息会被INFO和DEBUG处理器处理
        level = _get_level_num(self._logger_config["level"].upper())
        logging.basicConfig(
            level=level, stream=sys.stdout
        )

        # 创建主logger对象
        self._instance_logger = logging.getLogger('my_logger')
        self._instance_logger.propagate = False  # 禁止传播日志到根记录器, 必须有这个设置, 不然会有多个重复的打印

        # 获取langfuse日志对象
        langfuse_logger = logging.getLogger('langfuse')

        global_base_dir = self._global_config["global_base_dir"]
        log_dir = f"{global_base_dir}/logs"
        if not is_directory_exist(log_dir):
            create_directory(log_dir)

        # 创建不同日志级别的日志处理器和设置格式, 并添加到Logger对象
        for x_level in ["info", "error", "warning", "debug"]:
            log_file = f"{global_base_dir}/logs/{x_level}.{date}.log"
            handler = logging.FileHandler(log_file, encoding="utf-8")
            handler.setLevel(_get_level_num(x_level))
            handler.setFormatter(logging.Formatter(self._logger_config["log_format"]))
            self._instance_logger.addHandler(handler)
            langfuse_logger.addHandler(handler)

    def _log(self, level: int, msg: str, biz_data: dict, frame_data: dict):
        """
        内部底层写日志入口
        :param level: 日志级别
        :param msg: 消息内容
        :param biz_data: 业务数据
        :param frame_data: 框架数据
        """
        log_id = ""
        user_id = ""
        try:
            thread_local = threading.local()
            log_id = getattr(thread_local, "wpylib_threadkey_log_id")
            user_id = getattr(thread_local, "wpylib_threadkey_user_id")
        except Exception as e:
            ...

        # 日志模板
        template = convert_to_jsonable({
            # 业务信息
            "level": _get_level_text(level),
            "msg": msg,
            "log_id": log_id,
            "user_id": user_id,
            "data": biz_data,
            "project": self._global_config["project"],
            "business": self._global_config["business"],
            # 框架信息
            "thread_id": threading.current_thread().ident,
            "module": frame_data["module"] if "module" in frame_data else None,
            "caller": frame_data["caller"] if "caller" in frame_data else None,
            "client_ip": frame_data["client_ip"] if "client_ip" in frame_data else "127.0.0.1",
            "sysarg": sys.argv,
            # 时间信息
            "timestamp": now_timestamp(),
            "datetime": now_time_str(),
        })

        # 根据日志级别处理
        if level == logging.ERROR:
            # 封装日志模板信息
            template["traceback_info"] = traceback.format_exc()
            if template["traceback_info"] == "NoneType: None\n":
                template["traceback_info"] = ""
            template_string = f"ERROR: {json.dumps(template, ensure_ascii=False)}"

            # 记录ERROR记录
            self._instance_logger.error(template_string)

            # 自动在INFO日志中也冗余一份, 方便查询
            self._instance_logger.info(template_string)
        elif level == logging.INFO:
            # 封装日志模板信息
            template_string = f"INFO: {json.dumps(template, ensure_ascii=False)}"

            # 记录INFo日志
            self._instance_logger.info(template_string)
        else:
            raise RuntimeError(f"unknown log level")

        # 同步在输出缓存区
        print(template_string)

    def log_info(self, msg: str, biz_data: dict = None, frame_data: dict = None):
        """
        直接打印INFo级别日志
        :param msg: 消息内容
        :param biz_data: 业务数据字典结构
        :param frame_data: 框架数据
        """
        if biz_data is None:
            biz_data = {}
        if frame_data is None:
            frame_data = {}
        frame_data["module"] = _get_module(inspect.currentframe())
        frame_data["caller"] = _get_caller(inspect.stack())
        self._log(logging.INFO, f"[log_info]: {msg}", biz_data, frame_data)

    def log_error(self, msg: str, biz_data: dict = None, frame_data: dict = None):
        """
        直接打印ERROR级别日志
        :param msg: 消息内容
        :param biz_data: 业务数据字典结构
        :param frame_data: 框架数据
        """
        if biz_data is None:
            biz_data = {}
        if frame_data is None:
            frame_data = {}
        frame_data["module"] = _get_module(inspect.currentframe())
        frame_data["caller"] = _get_caller(inspect.stack())
        self._log(logging.ERROR, f"[log_error]: {msg}", biz_data, frame_data)


def _get_module(currentframe) -> str:
    """
    获取当前的module模块名
    :param currentframe: 当前帧
    :return: 返回模块名
    """
    try:
        return inspect.getmodule(currentframe).__name__
    except Exception as e:
        # 可能获取不到, 返回空即可, 不需要抛出异常
        return ""


def _get_caller(stack_list: list) -> str:
    """
    获取调用栈
    :param stack_list: 调用栈
    :return: 返回调用栈信息字符串
    """
    if len(stack_list) <= 1:
        return ""

    stack_2th = stack_list[1]
    if len(stack_2th) <= 3:
        return ""

    file = stack_2th[1]
    line = stack_2th[2]
    func = stack_2th[3]
    return f"{file}=>{line}=>{func}"


def _get_level_text(level: int) -> str:
    """
    获取日志级别(数字)所对应的字符串
    :param level: 日志级别(数字)
    :return: 返回日志级别(数字)所对应的字符串
    """
    level_text_map: dict = {
        logging.ERROR: "ERROR",
        logging.INFO: "INFO",
        logging.WARNING: "WARNING",
        logging.DEBUG: "DEBUG",
    }
    return level_text_map[level]


def _get_level_num(text: str) -> int:
    """
    获取日志级别(字符串)所对应的数字值
    :param text: 日志级别(字符串)
    :return: 返回日志级别(字符串)所对应的数字值
    """
    level_num_map: dict = {
        "ERROR": logging.ERROR,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "DEBUG": logging.DEBUG,
    }
    return level_num_map[text.upper()]



"""
@File: chain.py
@Date: 2024/12/10 10:00
@desc: 第三方chain模块
"""
from langchain import LLMChain
from langfuse.callback import CallbackHandler
from wpylib.pkg.langchain.log_callback import LogCallback
from langchain_core.prompts.base import BasePromptTemplate
from langchain_core.language_models.base import BaseLanguageModel
import threading


def get_chain_callbacks(langfuse_config: dict) -> list:
    """
    获取回调函数
    :return:
    """
    if "secret_key" not in langfuse_config:
        langfuse_config["secret_key"] = ""
    if "public_key" not in langfuse_config:
        langfuse_config["public_key"] = ""
    if "host" not in langfuse_config:
        langfuse_config["host"] = ""
    if "trace_name" not in langfuse_config:
        langfuse_config["trace_name"] = ""
    if "tags" not in langfuse_config:
        langfuse_config["tags"] = []

    thread_local = threading.local()
    log_id = getattr(thread_local, "wpylib_threadkey_log_id")

    callbacks = [
        CallbackHandler(
            secret_key=langfuse_config["secret_key"],
            public_key=langfuse_config["public_key"],
            host=langfuse_config["host"],
            debug=False,
            session_id=log_id,
            trace_name=langfuse_config["trace_name"],
            tags=langfuse_config["tags"],
        )
    ]
    return callbacks


def create_chain(model: BaseLanguageModel, prompt: BasePromptTemplate, verbose: bool = False) -> LLMChain:
    """
    创建chain
    chain.invoke(*args, config={"callbacks": get_callbacks()}, **kwargs)
    chain.predict(*args, callbacks=get_callbacks(), **kwargs)
    chain.apredict(*args, callbacks=get_callbacks(), **kwargs)
    :param model: 模型
    :param prompt: 提示词
    :param verbose: 是否开启打印
    :return:
    """
    return LLMChain(
        llm=model,
        prompt=prompt,
        verbose=verbose,
        callbacks=[LogCallback()]
    )



"""
@File: history.py
@Date: 2024/12/10 10:00
@desc: 第三方历史会话模块
"""
from wpylib.util.x.xtyping import is_none


def make_conversation_history(messages: list = None) -> str:
    """
    封装上下文
    :param messages: 消息记录
    :return:
    """
    conversation_history = ""
    if is_none(messages):
        messages = []
    for message in messages:
        conversation_history += f"\nUser: {message['query']}\nAssistant: {message['answer']}\n"
    return f"<ConversationHistory>{conversation_history}</ConversationHistory>"



"""
@File: http.py
@Date: 2024/12/10 10:00
@desc: HTTP请求工具
"""
from flask import request
from flask import jsonify
import requests

# HTTP请求方式
HTTP_METHOD_TYPE_GET: str = "get"
HTTP_METHOD_TYPE_POST: str = "post"

# 通用接口Code码
COMMON_HTTP_CODE_SUCCESS = 200
COMMON_HTTP_CODE_SYS_ERROR = 500
COMMON_HTTP_CODE_PARAMS_ERROR = 4001
COMMON_HTTP_CODE_LOGIN_AUTH_ERROR = 4002
COMMON_HTTP_CODE_MSG_MAP = {
    COMMON_HTTP_CODE_SUCCESS: "success",
    COMMON_HTTP_CODE_SYS_ERROR: "服务异常",
    COMMON_HTTP_CODE_PARAMS_ERROR: "参数校验失败",
    COMMON_HTTP_CODE_LOGIN_AUTH_ERROR: "用户未登录",
}


def get_params():
    """
    获取请求参数
    """
    post_params = {}
    url_params = request.args.to_dict()
    if request.method.lower() == HTTP_METHOD_TYPE_POST:
        if request.mimetype == "multipart/form-data":
            post_params = request.form
        elif request.mimetype == "application/json":
            post_params = request.json
    return url_params, post_params


def get_headers(header_name=""):
    """
    获取请求header
    """
    headers = request.headers
    if header_name != "":
        return headers[header_name] if header_name in headers else ""
    return headers


def requests_get(url: str, params=None, data=None, _json=None, parse_result=True, **kwargs):
    """
    使用requests库发起GET请求
    :param url: 请求地址
    :param params: 请求数据
    :param data: 请求数据, 类型为str
    :param _json: 请求数据, 类型为字典(即JSON), 即指的是一个可JSON序列化的Python对象
    :param parse_result: 是否解析结果
    :return: 返回请求结果
    """
    response = requests.get(url=url, params=params, data=data, json=_json, **kwargs)
    if response.status_code not in [0, 200]:
        raise RuntimeError(f"http code error: code={response.status_code} | response={response.text} | url={url}")

    if parse_result:
        try:
            res = response.json()
        except Exception as e:
            raise RuntimeError(f"requests_get exception: e={e}")
        return res
    return response.text


def resp_success(data):
    """
    响应成功
    :param data: 响应数据
    """
    res = {
        "code": 0,
        "data": data,
        "msg": "success",
    }
    return jsonify(res)


def resp_error(data=None, msg="操作失败", code=500):
    """
    响应错误
    """
    res = {
        "code": code,
        "data": data,
        "msg": msg,
    }
    return jsonify(res)


def resp_page_success(data_list, pg, pz, total):
    """
    返回分页数据
    :param data_list:
    :param pg:
    :param pz:
    :param total:
    :return:
    """
    return resp_success({
        'pg': pg,
        'pz': pz,
        'total': total,
        'list': data_list,
    })



"""
@File: stream_response.py
@Date: 2024/12/10 10:00
@Desc: 第三方流式响应模块
"""
from wpylib.util.x.xjson import stringify
from wpylib.pkg.sse.stream_queue import StreamQueue
from wpylib.util.http import COMMON_HTTP_CODE_SYS_ERROR
import threading
import queue

STREAM_MESSAGE_PING = "ping"


class StreamResponseGenerator(object):
    """
    流式响应生成器
    """

    def __init__(self, target, args=(), kwargs=None, queue_max: int = 20):
        if not kwargs:
            kwargs = {}
        self.queue = StreamQueue(queue_max)
        kwargs['queue'] = self.queue
        self.thread = threading.Thread(
            group=None,
            target=target,
            name='stream_response_generator_thread',
            args=args,
            kwargs=kwargs,
            daemon=True
        )

    def __iter__(self):
        self.thread.start()
        try:
            while self.queue.is_running():
                try:
                    e = self.queue.get(True, 3)
                except queue.Empty:
                    # websocket链接保活
                    self.queue.send_message(type_str=STREAM_MESSAGE_PING)
                    continue
                if not e:
                    continue
                if e["event"] == "message":
                    self.queue.task_done()
                    yield self._format_message(**e)
                    continue
                elif e["event"] == "message_end":
                    self.queue.task_done()
                    yield self._format_message(**e)
                    self.queue.close()
                    break
        except GeneratorExit:
            # 无法再yield
            # 此异常出现的时机为第三方sse断开连接导致无法yield, 该不可抗拒因素导致的异常无法处理，因此忽略该异常
            self.queue.close()
            self.thread.join(5)
            return
        except Exception as e:
            self.queue.send_message_end({"code": COMMON_HTTP_CODE_SYS_ERROR, "msg": "服务异常"})
            return
        finally:
            self.queue.close()
            self.thread.join(5)

    def _format_message(self, event, data):
        """
        格式化消息
        :param event: 事件
        :param data: 数据
        :return:
        """
        return "data: {}\n\n".format(stringify({
            "event": event,
            "data": data
        }))



"""
@File: storage.py
@Date: 2024/12/10 10:00
@desc: 本地存储工具
"""
import os


def is_file_exist(file) -> bool:
    """
    文件是否存在
    :param file: 文件路径
    :return 返回文件是否存在
    """
    return os.path.exists(file)


def is_directory_exist(directory: str) -> bool:
    """
    目录是否存在
    :param directory: 目录
    :return 返回目录是否存在
    """
    return os.path.exists(directory)


def create_directory(directory: str):
    """
    创建目录
    :param directory: 目录
    """
    os.mkdir(directory)



"""
@File: flask.py
@Date: 2024/12/10 10:00
@desc: 第三方flask单例模块
"""
from flask_cors import CORS
from flask.app import Flask as OfficialFlaskApp


class Flask:
    """
    二次封装的Flask类
    """
    # flask的app实例
    _instance_app: OfficialFlaskApp

    # Flask类的单例
    _singleton = None

    def __new__(cls, *args, **kwargs):
        # 在初始化Flask类的时候, 确保仅存在一个单例
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
        return cls._singleton

    def __init__(self, app_name: str):
        self._instance_app = OfficialFlaskApp(app_name)
        # 启用CORS，允许所有来源的跨域请求
        CORS(self._instance_app, supports_credentials=True)

    def get_instance_app(self):
        """
        返回flask的app实例
        :return:
        """
        return self._instance_app
